#include<linux/module.h>
#include<linux/slab.h>
#include<linux/mempool.h>
#include<linux/string.h>

#define MINIMUM_MEMPOOL_OBJECTS 10 

static struct example_struct {
	unsigned int id;
	char example_string[10];
} *example_struct_pointer;

static struct kmem_cache *example_cachep;
static mempool_t *mempool_pointer;

static void example_constructor(void *argument)
{
	static unsigned int id;
	static char test_string[] = "Test";
	struct example_struct *example = (struct example_struct *)argument;
	example->id = id;
	strcpy(example->example_string,test_string);
	id++;
}

void print_example_struct(struct example_struct *example)
{
	pr_notice("Example struct id: %u\n",example->id);
	pr_notice("Example string field content: %s\n",example->example_string);
}

static int __init slabmod_init(void)
{
	example_cachep = kmem_cache_create("example cache",sizeof(struct example_struct),0,SLAB_HWCACHE_ALIGN|SLAB_POISON|SLAB_RED_ZONE,example_constructor);
	if(IS_ERR(example_cachep)) {
		pr_alert("Error creating cache: %ld\n",PTR_ERR(example_cachep));
		return -ENOMEM;
	}

	mempool_pointer = mempool_create(MINIMUM_MEMPOOL_OBJECTS,mempool_alloc_slab,mempool_free_slab,example_cachep);
	if(IS_ERR(mempool_pointer)) {
		pr_alert("Error creating cache: %ld\n",PTR_ERR(mempool_pointer));
		kmem_cache_destroy(example_cachep);
		return -1;
	}

	example_struct_pointer = (struct example_struct *)mempool_alloc(mempool_pointer,GFP_KERNEL);
	if(IS_ERR(example_struct_pointer)) {
		pr_alert("Error allocating form cache: %ld\n",PTR_ERR(example_struct_pointer));
		mempool_destroy(mempool_pointer);
		kmem_cache_destroy(example_cachep);
		return -ENOMEM;
	}
	
	return 0;
}

static void __exit slabmod_exit(void)
{
	if(example_cachep) {
		if(mempool_pointer) {
			if(example_struct_pointer) {
				print_example_struct(example_struct_pointer);
				mempool_free(example_struct_pointer,mempool_pointer);
			}
			mempool_destroy(mempool_pointer);
		}
		kmem_cache_destroy(example_cachep);
	}
}

module_init(slabmod_init);
module_exit(slabmod_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Arkadiusz Chrobot <a.chrobot@tu.kielce.pl>");
MODULE_DESCRIPTION("A module demonstrating the use of a mempool.");
MODULE_VERSION("1.0");
