#include<linux/module.h>
#include<linux/gfp.h>       
#include<linux/mm.h>        

#define ORDER 2

static struct page *page_pointer, *pages_pointer;
static unsigned long int single_page_address, zeroed_page_address, multiple_pages_address;


static int __init mallocmod_init(void)
{
	pages_pointer = alloc_pages(GFP_KERNEL,ORDER);
	if(IS_ERR(pages_pointer)) {
		pr_alert("Error allocating %u pages: %ld!\n",1<<ORDER, PTR_ERR(pages_pointer));
		return -ENOMEM;
	}
	pr_notice("Pages address: %p\n",page_address(pages_pointer));

	page_pointer = alloc_page(GFP_KERNEL);
	if(IS_ERR(page_pointer)) {
		pr_alert("Error allocating page: %ld!\n",PTR_ERR(page_pointer));
		return -ENOMEM;
	}
	pr_notice("Pages address: %p\n",page_address(page_pointer));
	
	multiple_pages_address = __get_free_pages(GFP_KERNEL,ORDER);
	if(!multiple_pages_address) {
		pr_alert("Error allocating %u pages!\n",1<<ORDER);
		return -ENOMEM;
	}
	pr_notice("Pages address: %lx\n",multiple_pages_address);
	
	single_page_address = __get_free_page(GFP_KERNEL);
	if(!single_page_address) {
		pr_alert("Error allocating page!\n");
		return -ENOMEM;
	}
	pr_notice("Page address: %lx\n",single_page_address);

	zeroed_page_address = get_zeroed_page(GFP_KERNEL);
	if(!zeroed_page_address) {
		pr_alert("Error allocating zeroed page!\n");
		return -ENOMEM;
	}
	pr_notice("Page address: %lx\n",zeroed_page_address);

	return 0;
}

static void __exit mallocmod_exit(void)
{
	if(zeroed_page_address)
		free_page(zeroed_page_address);
	if(single_page_address)
		free_page(single_page_address);
	if(multiple_pages_address)
		free_pages(multiple_pages_address,ORDER); 
	if(page_pointer) 
		__free_page(page_pointer);
	if(pages_pointer) 
		__free_pages(pages_pointer,ORDER);
}

module_init(mallocmod_init);
module_exit(mallocmod_exit);

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Arkadiusz Chrobot <a.chrobot@tu.kielce.pl>");
MODULE_DESCRIPTION("A module demonstrating the usage of zone allocator.");
MODULE_VERSION("1.0");
